from ejabberd_python3d.abc.api import Enum


class LogLevelOptions(Enum):
    none = 1
    emergency = 2
    alert = 3
    critical = 4
    error = 5
    warning = 6
    notice = 7
    info = 8
    debug = 9
